#!/usr/bin/env python2

import os
import glob
import time
import datetime
import sys


def formatPrint(text, *argv):

    separator = '>>'
    now = datetime.datetime.now().strftime('%Y %j %H:%M:%S')
    message = "%s %s %s" % (now, separator, text)
    message += " ".join([str(arg) for arg in argv])
    print message



# Check if it is being run on another process
pid = str(os.getpid())
pidfile = "/var/run/logmonitor.pid"
if os.path.isfile(pidfile):
    #formatPrint("%s already exists, checking if there is a process running." % pidfile)
    command = "ps up $(cat %s) | grep monitor.py" % pidfile
    #formatPrint("Running: %s" % command)
    ret = os.system(command)
    #formatPrint("Returns: ", ret)
    if not ret:
        #formatPrint("There was a process already running!")
        sys.exit()

formatPrint("Starting new process.")

file(pidfile, 'w').write(pid)


# collection of paths (maybe wildcarded) to log files
PATHS = [
    '/opt/log/*',
    '/opt/seiscomp/acquisition/log/seedlink.log'
]

EXCEPTIONS = [
    '.old',
    'post-boot.log'
]

# if disk usage gets over this amount (%), log files are not examined
# on line by line basis, but truncated to MAX_SIZE length (preserving
# the latest part of information).
DISK_LIM  = 70          # 70%
MAX_SIZE  = 5 *1024 * 1024 # 5M

# in normal condition (when there's no disk alert) the program deletes
# lines from timestamped logs which are older than this amount of days
DAYS_KEEP = 7

# the program sleeps for this amount of seconds
SLEEP_FOR = 5 *60 * 60 # 5 hours.

#########################################################################

# Expand wildcards in PATHS

NAMES = []
for path in PATHS:
    for name in glob.glob(path):
        NAMES.append(name)

formatPrint('')
formatPrint('\tFiles to explore:')
formatPrint('')
for name in NAMES:
    formatPrint(name)
formatPrint('')

#########################################################################

# Timestamp decoders

def time0(line):

    toks = line.split('>>')
    if len(toks) < 2:
        return None

    try:
        return datetime.datetime.strptime(toks[0], '%Y %j %H:%M:%S')
    except:
        pass

    return None


def time1(line):

    toks = line.split()
    if len(toks) < 3:
        return None

    if toks[2][-1] != ':':
        return None

    s = ' '.join(toks[:3])
    try:
        return datetime.datetime.strptime(s, '%Y %j %H:%M:%S:')
    except:
        pass

    return None


def time2(line):

    toks = line.split('-')
    if len(toks) < 2:
        return None

    if toks[0][-1] != ' ':
        return None

    try:
        return datetime.datetime.strptime(toks[0], '%a %b %d %H:%M:%S %Y ')
    except:
        pass

    return None

# searches a file for a line which can be interpreted as
# a timestamp decodable with any of the above functions,
# and returns corresponding function.

def select_time_decoder(name):
    try:
        with open(name, 'r') as f:
            for line in f:
                if time0(line) is not None:
                    return time0

                if time1(line) is not None:
                    return time1

                if time2(line) is not None:
                    return time2
    except:
        formatPrint('Cannot read %s' % name)

    return None

#########################################################################

def disk_usage():
    p = os.popen('''df -h | awk '{ if ($6 == "/") print $5 " used, " $4 " available" }' ''')
    l = p.readline().strip()
    formatPrint('Disk usage: %s' % l)
    usage = l.split()[0].replace('%','')
    usage = float(usage)
    return usage

#########################################################################

# Get file size with the proper format

def formatedFileSize(name):
    rawSize = getSize(name)
    return formatSize(rawSize)

def formatSize(size):
    kb = size/1024

    if kb < 1024:
        return "%d kB" % kb
    else:
        return "%.2f MB" % (kb/1024.)

def getSize(name):
    return os.path.getsize(name)


# truncate file to size

def trunc_size(name, size):

    flen = getSize(name)
    formatPrint('- Old file size %s' % formatedFileSize(name))

    if flen <= size:
        formatPrint('- No need to truncate')
        return

    formatPrint('- Truncating file to %s' % formatSize(flen))

    data = None

    with open(name, 'r') as ifile:
        ifile.seek(flen - size)
        data = ifile.read()

    with open(name, 'w') as ofile:
        ofile.write(data)

    formatPrint('- New file size %s' % formatedFileSize(name))

# remove obsolete records

def trunc_date(name, date):

    formatPrint('- Old file size %s' % formatedFileSize(name))

    time_func = select_time_decoder(name)

    if time_func is None:
        formatPrint('- Could not pick a timestamp decoder, return')
        return

    flag = False
    save = []
    needToModify = True

    tot = 0
    obs = 0

    with open(name, 'r') as ifile:
        for line in ifile:

            tot += 1

            if flag:
                save.append(line)
                continue

            newd = time_func(line)
            if newd is not None and newd >= date:
                flag = True
                if tot == 1:
                    needToModify = False
                    break
                continue

            obs += 1

    if needToModify:
        formatPrint('- Truncating %d lines of %d' % (obs, tot))

        with open(name, 'w') as ofile:
            ofile.writelines(save)

        formatPrint('- New file size %s' % formatedFileSize(name))

#########################################################################

def explore_logs():

    low_disk = disk_usage() > DISK_LIM

    if low_disk:
        formatPrint('--- Insufficient disk space condition.')
        formatPrint('--- Files are truncated to %s' % formatSize(MAX_SIZE))

    old_date = datetime.datetime.utcnow() - datetime.timedelta(DAYS_KEEP)

    for name in NAMES:
        isException = False
        for pattern in EXCEPTIONS:
            if pattern in name:
                isException = True
        if isException:
            continue

        formatPrint('')
        formatPrint('Examining file %s' % name)

        if not os.path.isfile(name):
            formatPrint('- %s is not a file' % name)
            continue

        if low_disk:
            formatPrint('- Truncating files to %s' % formatSize(MAX_SIZE))
            trunc_size(name, MAX_SIZE)
        else:
            formatPrint('- Deleting lines older than %d days' % DAYS_KEEP)
            trunc_date(name, old_date)


while True:
    explore_logs()
    formatPrint('')
    formatPrint('---')
    formatPrint('--- UTC %s' % datetime.datetime.utcnow().strftime('%Y %m %d %H:%M:%S'))
    formatPrint('--- Sleeping for %d seconds' % SLEEP_FOR)
    formatPrint('---')
    formatPrint('')
    time.sleep(SLEEP_FOR)



os.remove(pidfile)
